#include <libce.h>
#include <vm.h>

// modifichiamo vm1 in modo da permettere la scrittura tramite vx in un
// primo momento e poi vietarla in un secondo momento
int x = 0;

// Nota: nelle intenzioni originarie, il bit R/W=0 serve solo ad impedire le
// scritture *da livello utente* e viene di conseguenza ignorato se il processore
// si trova a livello sistema.  Opzionalmente, però, il codice di sistema può
// chiedere alla MMU di controllare il permesso di scrittura anche per il
// codice di sistema stesso (questo può essere utile, per esempio, per
// intercettare eventuali bug nel codice di sistema). Dal momento che questo
// programma girerà a livello sistema, dobbiamo abilitare questa opzione se
// vogliamo sperimentare gli effetti dei bit R/W. L'opzione si abilita settando
// il bit 16 del registro cr0. Lo facciamo fare alla funzione
// enable_write_protect(), scritta in vm2.s:
extern "C" void enable_write_protect();
// Il modulo sistema del nucleo abilita questa opzione, appunto per aiutare ad
// intercettare errori quando si risolvono gli esercizi.

void main()
{
	enable_write_protect();
	// copiamo il codice di vm1, ma settiamo anche tutti i bit R/W
	vaddr v = norm(04443332221110000);
	paddr px = trasforma(readCR3(), int_cast<vaddr>(&x));
	paddr base_px = px & ~0xFFF;
	paddr tab4 = readCR3();
	paddr tab3 = alloca_tab();
	paddr tab2 = alloca_tab();
	paddr tab1 = alloca_tab();
	tab_entry& e4 = get_entry(tab4, 0444);
	tab_entry& e3 = get_entry(tab3, 0333);
	tab_entry& e2 = get_entry(tab2, 0222);
	tab_entry& e1 = get_entry(tab1, 0111);

	e4 = tab3    | BIT_P | BIT_RW;
	e3 = tab2    | BIT_P | BIT_RW;
	e2 = tab1    | BIT_P | BIT_RW;
	e1 = base_px | BIT_P | BIT_RW;

	natq offset_px = px & 0xFFF;
	vaddr vx = v | offset_px;
	int *p = ptr_cast<int>(vx);

	// questo accesso dovrebbe essere consentito (non dovrebbe generare
	// page fault)
	*p = 1;
	// ora resettiamo il bit RW da una entrata qualsiasi
	e1 &= ~BIT_RW;
	// in teoria questo accesso dovrebbe generare page fault, ma
	// molto probabilmente sarà invece eseguito correttamente. Come mai?
	*p = 2;
	// Il motivo è che, dopo il primo accesso, la traduzione si trova
	// ormai nel TLB e, per il secondo accesso, la MMU userà quella senza
	// ripercorrere il TRIE e, dunque, senza accorgersi del cambiamento
	// del bit RW. Per rendere effettiva la modifica dobbiamo quindi
	// invalidare il TLB. Non c'è bisogno però di invalidarlo tutto:
	// usiamo l'istruzione invlpg per invalidare soltanto la traduzione
	// di vx (o, più precisamente, della pagina [v, v+4096)).
	invalida_entrata_TLB(vx); // equivalentemente: invalida_entrata_TLB(v)
	// ora questo accesso causerà page fault
	*p = 3;

	pause();

}
