#include <libce.h>
#include <vm.h>

// Usiamo un tab_iter per stampare lo stato delle entrate lungo un percorso del
// TRIE. Riusiamo parte del codice di vm2.
int x = 0;

extern "C" void enable_write_protect();

void main()
{
	enable_write_protect();

	vaddr v = norm(04443332221110000);

	paddr px = trasforma(readCR3(), int_cast<vaddr>(&x));

	paddr base_px = px & ~0xFFF;

	paddr tab4 = readCR3();
	paddr tab3 = alloca_tab();
	paddr tab2 = alloca_tab();
	paddr tab1 = alloca_tab();

	tab_entry& e4 = get_entry(tab4, 0444);
	tab_entry& e3 = get_entry(tab3, 0333);
	tab_entry& e2 = get_entry(tab2, 0222);
	tab_entry& e1 = get_entry(tab1, 0111);

	e4 = tab3    | BIT_P | BIT_RW;
	e3 = tab2    | BIT_P | BIT_RW;
	e2 = tab1    | BIT_P | BIT_RW;
	e1 = base_px | BIT_P | BIT_RW;

	natq offset_px = px & 0xFFF;
	vaddr vx = v | offset_px;

	int *p = ptr_cast<int>(vx);

	// Prima del primo accesso dovremmo trovare le entrate esattamente
	// nello stato in cui le abbiamo preparate (assegnamenti a e4-e1 qui
	// sopra). Nella rappresentazione esadecimale di una entrata i bit P,
	// RW etc. sono tutti nel byte meno significativo, e dunque sono tutti
	// nelle due cifre esadecimali meno significative. Il bit P è il bit 0
	// e il bit RW è il bit 1, quindi la cifra meno significativa di ogni
	// entrata dovrebbe essere 3 (11 in binario)
	printf("Prima del primo accesso:\n");
	for (tab_iter it(tab4, vx); it; it.next()) {
		int l = it.get_l();
		tab_entry e = it.get_e();
		paddr tab = it.get_tab();
		printf("livello %d, tabella %6lx, entrata %6lx\n", l, tab, e);
		// Nota: Sia paddr che tab_entry sono unsigned long.
		// `lx` è il placeholder per long (la `l`) unsigned in
		// esadecimale (la `x`).
		// Il numero `6` prima di `lx` serve a specificare la larghezza
		// del campo, in modo che printf allinei le varie stampe
	}
	volatile int y = *p;
	// A causa di questo accesso, la MMU dovrebbe aver settato i bit A
	// lungo tutto il percorso. Il bit A è il numero 5, quindi dovremmo
	// osservare
	//
	// 	7 6 5 4|3 2 1 0		(posizione dei bit)
	// 	---------------
	// 	0 0 1 0|0 0 1 1		(binario)
	//	---------------
	//	   2   |   3  		(esadecimale)
	printf("Dopo il primo accesso:\n");
	for (tab_iter it(tab4, vx); it; it.next()) {
		int l = it.get_l();
		tab_entry e = it.get_e();
		paddr tab = it.get_tab();
		printf("livello %d, tabella %6lx, entrata %6lx\n", l, tab, e);
	}
	*p = 1;
	// A causa di questo accesso la MMU dovrebbe aver settato il bit D
	// Nell'entrata di livello 1. Il bit D è il numero 6, quindi dovremmo
	// osservare (solo nel livello 1)
	//
	// 	7 6 5 4|3 2 1 0		(posizione dei bit)
	// 	---------------
	// 	0 1 1 0|0 0 1 1		(binario)
	//	---------------
	//	   6   |   3  		(esadecimale)
	printf("Dopo l'accesso in scrittura:\n");
	for (tab_iter it(tab4, vx); it; it.next()) {
		int l = it.get_l();
		tab_entry e = it.get_e();
		paddr tab = it.get_tab();
		printf("livello %d, tabella %6lx, entrata %6lx\n", l, tab, e);
	}

	pause();

}
