#include <libce.h>
#include <vm.h>

// Oltre che una funzione, map() può usare un "funtore", cioè
// un oggetto che definisce operator() e dunque può essere usato
// come una funzione.
//
// Come esempio proviamo a creare nuove traduzioni per 10 pagine
// fisiche a partire dall'indirizzo 0x200000.
static const int NPAGES = 10;

// L'oggetto ricorda a quale pagina siamo arrivati (variabile i)
// e restituisce il prossimo indirizzo fisico ogni volta che viene
// invocato operator(), incrementando i per la prossima volta.
struct functor {
	paddr base;
	int i;

	functor(paddr base_): base(base_), i(0) {}

	paddr operator() (vaddr) {
		paddr p = base + i * DIM_PAGINA;
		i++;
		return p;
	}
};

void main()
{
	vaddr v = norm(04443332221110000);
	paddr tab4 = readCR3();

	functor f(0x20000);
	// Usiamo in tab_iter prima e dopo l'invocazione di map(), per
	// mostrare cosa fa la fuzione.
	printf("==> Prima di map():\n");
	for (tab_iter it(tab4, v, NPAGES * DIM_PAGINA); it; it.next()) {
		// nota: `o` stampa in ottale
		printf("indirizzo %16lo, livello %d, tabella %6lx, entrata %6lx\n",
				it.get_v(), it.get_l(), it.get_tab(), it.get_e());
	}

	map(tab4, v, v + NPAGES * DIM_PAGINA, BIT_RW, f);

	printf("==> Dopo map():\n");
	for (tab_iter it(tab4, v, NPAGES * DIM_PAGINA); it; it.next()) {
		printf("indirizzo %16lo, livello %d, tabella %6lx, entrata %6lx\n",
				it.get_v(), it.get_l(), it.get_tab(), it.get_e());
	}

	pause();

}
