#include <libce.h>
#include <vm.h>

// Il modo più comodo di usare map() prevede l'uso delle "espressioni lambda",
// che permettono di definire funzioni senza nome nel mezzo di altre funzioni.
//
// La sintassi (semplificata) è:
//
// 	[] (argomenti) -> tipo-di-ritorno { codice }
//
// Dove "-> tipo-di-ritorno" si può anche omettere.

static const int NPAGES = 4;

void main()
{
	paddr tab4 = readCR3();

	vaddr v1 = norm(0xaaaaaaaaa000);
	vaddr v2 = v1 + NPAGES/2 * DIM_PAGINA;

	printf("==> Prima delle map():\n");
	for (tab_iter it(tab4, v1, NPAGES * DIM_PAGINA); it; it.next()) {
		printf("indirizzo %16lx, livello %d, tabella %6lx, entrata %6lx\n",
				it.get_v(), it.get_l(), it.get_tab(), it.get_e());
	}

	// mappiamo due pagine a partire da v1 in se stesse
	map(tab4, v1, v2, BIT_RW, [] (vaddr v) { return v; });
	printf("==> Dopo la prima map():\n");
	for (tab_iter it(tab4, v1, NPAGES * DIM_PAGINA); it; it.next()) {
		printf("indirizzo %16lx, livello %d, tabella %6lx, entrata %6lx\n",
				it.get_v(), it.get_l(), it.get_tab(), it.get_e());
	}
	// Ora mappiamo due pagine a partire da v2 negli indirizzi fisici a
	// partire da 0x200000. Per far questo la lambda deve contare quante
	// volte è stata chiamata. Possiamo usare una variable locale definita
	// in questa stessa funzione e fare in modo che la lambda vi acceda per
	// riferimento. Per permettere quest'ultima cosa scriviamo `[&]`
	// invece di `[]`:
	int i = 0; // conta quante volte viene invocata la lambda
	map(tab4, v2, v2 + NPAGES/2 * DIM_PAGINA, BIT_RW,
		[&] (vaddr) {
			paddr p = 0x20000 + i * DIM_PAGINA;
			i++; // incrementa la variabile i di main
			return p;
		});
	// ora i vale NPAGES/2

	printf("==> Dopo la seconda map():\n");
	for (tab_iter it(tab4, v1, NPAGES * DIM_PAGINA); it; it.next()) {
		printf("indirizzo %16lx, livello %d, tabella %6lx, entrata %6lx\n",
				it.get_v(), it.get_l(), it.get_tab(), it.get_e());
	}

	pause();
}
